import math
from decimal import Decimal
from fractions import Fraction

def fexp(number):
    (sign, digits, exponent) = Decimal(number).as_tuple()
    return len(digits) + exponent - 1

def fman(number):
    return Decimal(number).scaleb(-fexp(number)).normalize()

def frac_approx(x,max_denom):
    frac = Fraction(x).limit_denominator(max_denom)
    return frac.numerator,frac.denominator

def format_output(result,mode):
    mant = fman(result)
    expo = fexp(result)
    if mode == "sci":
        return "%f*10^%i"%(mant,expo)
    if mode == "eng":
        ecor = expo%3
        return "%f*10^%i"%(10**ecor*mant,expo-ecor)
    if mode == "auto":
        return str(result)
    if mode == "dec":
        return "not supported"
    if mode == "frac":
        return "%i/%i"%frac_approx(result,10000)




from pygments.lexer import RegexLexer, words
from pygments.token import *

from . import functions

class HighlightLexer(RegexLexer):
    name = 'HighlightLexer'
    aliases = [ 'highlightlexer' ]

    tokens = {
        'root': [
            ('^#.*$', Comment.Preproc),
            (r'\b(and|or|not|mod)\b', Operator),
            (words((functions.get_builtin_names()), suffix=r'\b'), Name.Builtin),
            ('[a-zA-Z_]\w*', Name),
            (r'(\d+\.\d*|\d+)([eE][+-]?[0-9]+)?', Number),
            (r'\d+', Number),
            (r'\(|\)|\[|\]|\{|\}|,', Punctuation),
            (r'=|\+|-|\*|/|\^|>|<|>=|<=|==|\!=|@|\$|->', Operator),
        ]
    }



from prompt_toolkit.shortcuts import prompt,clear,print_tokens
from prompt_toolkit.styles import style_from_dict
from prompt_toolkit.token import Token
from prompt_toolkit.layout.lexers import PygmentsLexer
from prompt_toolkit.history import FileHistory
from prompt_toolkit.contrib.completers import WordCompleter
from os.path import expanduser

from .highlightlexer import HighlightLexer
from . import functions
from .lexer import *
from .parser import Parser,FuncDefAST,VarDefAST
from .runtime import BcEnvironment
from . import output
from .errors import err_msg_tokens,CommandError

calc_style = style_from_dict({
    Token: '#ffffff',
    Token.Comment.Preproc: '#ff5f00',
    Token.Operator: '#87ff5f',
    Token.Punctuation: '#87ff5f',
    Token.Name.Builtin: '#5fffff',
    Token.Prompt: '#ffff00 bold',
    Token.Number: '#ff5fff',
    Token.ErrorBold: '#ff0000 bold',
    Token.Error: '#ff0000'
})

config = {
    "vi_mode": False,
    "out_mode": "auto"
}

def get_ptok_main(num):
    def tok_func(cli):
        return [
            (Token.Prompt, '['),
            (Token.Number, str(num)),
            (Token.Prompt, ']: ')
        ]
    return tok_func

def get_ptok_continue(cli):
    return [
        (Token.Prompt, '.. ')
    ]

def index_split(str):
    tokens = []
    indices = []

    in_tok = False
    for i,c in enumerate(str):
        if not in_tok:
            if not c.isspace():
                tokens.append(c)
                indices.append(i)
                in_tok = True
        else:
            if c.isspace():
                in_tok = False
            else:
                tokens[-1] += c

    return tokens,indices

def main():
    history = FileHistory(expanduser("~/.gkraz_history"))
    completer = WordCompleter(functions.get_builtin_names())

    env = BcEnvironment()

    while True:
        try:
            is_reset = False
            input = prompt(
                    get_prompt_tokens=get_ptok_main(env.get_num()),
                    style=calc_style,
                    lexer=PygmentsLexer(HighlightLexer),
                    vi_mode=config["vi_mode"],
                    completer=completer,
                    complete_while_typing=False,
                    history=history)
            text = input.strip()
            if text != '':
                if text[0] == '#':
                    cmd,indices = index_split(text[1:])
                    if len(cmd) == 0:
                        raise CommandError("Expected command after '#'",1)
                    elif cmd[0] == "exit" or cmd[0] == "quit":
                        if len(cmd) != 1:
                            raise CommandError(cmd[0] + " expects no arguments",indices[1]+1)
                        break
                    elif cmd[0] == "mode":
                        if len(cmd) == 2:
                            if (cmd[1] == "vi"):
                                config["vi_mode"] = True
                            elif (cmd[1] == "emacs"):
                                config["vi_mode"] = False
                            elif (cmd[1] == "sci"):
                                config["out_mode"] = "sci"
                            elif (cmd[1] == "auto"):
                                config["out_mode"] = "auto"
                            elif (cmd[1] == "eng"):
                                config["out_mode"] = "eng"
                            elif (cmd[1] == "dec"):
                                config["out_mode"] = "dec"
                            elif (cmd[1] == "frac"):
                                config["out_mode"] = "frac"
                            else:
                                raise CommandError("Invalid argument for mode",indices[1]+1)
                        else:
                            raise CommandError("mode expects one argument",indices[2]+1)
                    elif cmd[0] == "reset":
                        if len(cmd) != 1:
                            raise CommandError("reset expects no arguments",indices[1]+1)
                        env = BcEnvironment()
                        is_reset = True
                    elif cmd[0] == "clear":
                        if len(cmd) != 1:
                            raise CommandError("clear expects no arguments",indices[1]+1)
                        clear()
                    elif cmd[0] == "matrix":
                        if len(cmd) == 2:
                            input = []
                            while True:
                                input.append(prompt(
                                        get_prompt_tokens=get_ptok_continue,
                                        style=calc_style,
                                        lexer=PygmentsLexer(HighlightLexer),
                                        vi_mode=config["vi_mode"],
                                        completer=completer,
                                        complete_while_typing=False))
                                if input[-1].strip() == "":
                                    break
                            mat = [line.split(" ") for line in input[:-1]]
                            if len(mat) == 0:
                                raise Exception("empty matrix")
                            lines = [ "[" + ",".join(line) + "]" for line in mat ]
                            code  = cmd[1] + " = [" + ",".join(lines) + "]"
                            print(code)

                            parser = Parser(lex(code))
                            stmt = parser.parseStmt()
                            bc = []
                            stmt.gen_bc(bc)
                            env.add_var(stmt.name,bc)
                        else:
                            raise CommandError("matrix expects variable name as argument",
                                    indices[1]+1)
                    else:
                        raise CommandError("unknown command",1)

                else:
                    parser = Parser(lex(text))
                    stmt = parser.parseStmt()
                    bc = []
                    stmt.gen_bc(bc)

                    if isinstance(stmt,FuncDefAST):
                        env.add_func(stmt.name,stmt.params,bc)
                    elif isinstance(stmt,VarDefAST):
                        env.add_var(stmt.name,bc)
                    else:
                        result = env.run_expr(bc)
                        if isinstance(result,float):
                            print(output.format_output(result,config["out_mode"]))
                        elif result is not None:
                            print(result)
            if not is_reset: env.inc_num()
        except KeyboardInterrupt:
            pass
        except Exception as ex:
            env.inc_num()
            print_tokens(err_msg_tokens(ex,env.get_num()),style=calc_style)

if __name__ == "__main__":
    main()



from prompt_toolkit.token import Token

class StmtException(Exception):
    def __init__(self,message,pos):
        self.message = message
        self.pos = pos

    def __str__(self):
        return self.message

class ParseError(StmtException):
    pass

class CommandError(StmtException):
    pass

def err_msg_tokens(err,num):
    if isinstance(err,StmtException):
        postok = [(Token.Prompt," "*(len(str(num))+4+err.pos) + "^\n")]
    else:
        postok = []


    return postok + [
        (Token.ErrorBold, type(err).__name__ + ": "),
        (Token.Error, str(err)+"\n")
    ]



from enum import Enum

class OpCo(Enum):
    ADD =  1
    SUB =  2
    MUL =  3
    DIV =  4
    MOD =  5
    AND =  6
    OR  =  7
    NOT =  8
    NEG =  9
    EQ  = 10
    NEQ = 11
    GT  = 12
    GTE = 13
    LT  = 14
    LTE = 15
    POW = 16
    FUN = 17
    VAR = 18
    LIT = 19
    VEC = 20
    STO = 21
    LD  = 22
    DOT = 23
    UNI = 24
    TO  = 25

class Bytecode:
    def __init__(self,opcode,*args):
        self.opcode = opcode
        self.args = args

    def print_bc(self):
        print(self.opcode.name,self.args)



import numpy as np
from pint import UnitRegistry

from .bytecode import *
from .functions import FUNCS,CONSTS

class BcEnvironment:
    def __init__(self):
        self.globs = {}
        self.num = 1
        self.ureg = UnitRegistry()

    def get_num(self):
        return self.num

    def inc_num(self):
        self.num += 1

    def add_func(self,name,params,bc):
        def newfunc(*args):
            if len(args) != len(params):
                raise TypeError("%s() takes %i positional argument but %s were given"%
                        (name,len(params),len(args)))
            locals = { param: arg for param,arg in zip(params,args) }
            return self.run_bc(bc,locals)
        self.globs[name] = newfunc

    def add_var(self,name,bc):
        self.globs[name] = self.run_bc(bc)

    def run_expr(self,bc):
        return self.run_bc(bc)

    def run_bc(self,bc,locals={}):
        stack = []
        storage = {}

        def get_var(name,var):
            if name in locals:
                return locals[name]
            if name in self.globs:
                return self.globs[name]
            if name in FUNCS:
                return FUNCS[name]
            if name in CONSTS:
                return CONSTS[name]
            raise NameError("%s %s not defined"%("Variable" if var else "Function",name))

        def o_add(opargs):
            stack.append(stack.pop()+stack.pop())
        def o_sub(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2-s1)
        def o_mul(opargs):
            stack.append(stack.pop()*stack.pop())
        def o_div(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2/s1)
        def o_mod(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2%s1)
        def o_and(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s1 and s2)
        def o_or(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s1 or s2)
        def o_not(opargs):
            stack.append(not stack.pop())
        def o_neg(opargs):
            stack.append(-stack.pop())
        def o_eq(opargs):
            stack.append(stack.pop()==stack.pop())
        def o_neq(opargs):
            stack.append(stack.pop()!=stack.pop())
        def o_gt(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2>s1)
        def o_gte(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2>=s1)
        def o_lt(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2<s1)
        def o_lte(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2<=s1)
        def o_pow(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2**s1)
        def o_fun(opargs):
            args = stack[-opargs[1] or None:]
            del stack[-opargs[1] or None:]
            stack.append(get_var(opargs[0],False)(*args))
        def o_var(opargs):
            stack.append(get_var(opargs[0],True))
        def o_lit(opargs):
            stack.append(opargs[0])
        def o_vec(opargs):
            vec = np.array(stack[-opargs[0] or None:])
            del stack[-opargs[0] or None:]
            stack.append(vec)
        def o_sto(opargs):
            storage[opargs[0]] = stack.pop()
        def o_ld(opargs):
            stack.append(storage[opargs[0]])
        def o_dot(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2 @ s1)
        def o_uni(opargs):
            stack.append(self.ureg.parse_expression(opargs[0]))
        def o_to(opargs):
            s1 = stack.pop()
            s2 = stack.pop()
            stack.append(s2.to(s1))

        odict = {
            OpCo.ADD: o_add,
            OpCo.SUB: o_sub,
            OpCo.MUL: o_mul,
            OpCo.DIV: o_div,
            OpCo.MOD: o_mod,
            OpCo.AND: o_and,
            OpCo.OR : o_or ,
            OpCo.NOT: o_not,
            OpCo.NEG: o_neg,
            OpCo.EQ : o_eq ,
            OpCo.NEQ: o_neq,
            OpCo.GT : o_gt ,
            OpCo.GTE: o_gte,
            OpCo.LT : o_lt ,
            OpCo.LTE: o_lte,
            OpCo.POW: o_pow,
            OpCo.FUN: o_fun,
            OpCo.VAR: o_var,
            OpCo.LIT: o_lit,
            OpCo.VEC: o_vec,
            OpCo.STO: o_sto,
            OpCo.LD : o_ld ,
            OpCo.DOT: o_dot,
            OpCo.UNI: o_uni,
            OpCo.TO : o_to
        }
        for op in bc:
            odict[op.opcode](op.args)
        self.globs["_%i"%self.num] = stack[0]
        if "__" in self.globs: self.globs["___"] = self.globs["__"]
        if "_" in self.globs: self.globs["__"] = self.globs["_"]
        self.globs["_"] = stack[0]
        return stack[0]



import numpy as np
import matplotlib.pyplot as plt
from terminaltables import SingleTable

def g_plot(x,y):
    plt.plot(x,y)
    plt.show()

def g_table(*cols):
    lines = list(zip(*cols))
    tbl = SingleTable(lines)
    tbl.inner_heading_row_border = False;
    return tbl.table

CONSTS = {
    "pi": np.pi,
    "e": np.e,
    "i": 1j
}

FUNCS = {
    "sqrt": np.sqrt,
    "sin": np.sin,
    "cos": np.cos,
    "tan": np.tan,
    "arcsin": np.arcsin,
    "arccos": np.arccos,
    "arctan": np.arctan,
    "atan2": np.arctan2,
    "arctan2": np.arctan2,
    "exp": np.exp,
    "ln": np.log,
    "log": np.log,
    "log2": np.log2,
    "log10": np.log10,
    "logb": lambda x,b: np.log(x)/np.log(b),
    "sinh": np.sinh,
    "cosh": np.cosh,
    "tanh": np.tanh,
    "asinh": np.arcsinh,
    "acosh": np.arccosh,
    "atanh": np.arctanh,
    "arcsinh": np.arcsinh,
    "arccosh": np.arccosh,
    "arctanh": np.arctanh,
    "if": np.where,
    "step": lambda x: np.heaviside(x,0),
    "rect": lambda x: np.heaviside(0.5+x,0)*np.heaviside(0.5-x,0),

    "arg": np.angle,
    "real": np.real,
    "imag": np.imag,
    "re": np.real,
    "im": np.imag,
    "conj": np.conj,

    "range": np.arange,
    "linspace": np.linspace,

    "plot": g_plot,
    "table": g_table,

    "magn": lambda x: x.magnitude,
    "dim": lambda x: x.dimensionality,
    "unit": lambda x: x.units
}
def get_builtin_names():
    return list(FUNCS.keys()) + list(CONSTS.keys())



from .errors import ParseError

class LexTok:
    ID  = 1
    SYM = 2
    NUM = 3
    EOF = 4
    def __init__(self,type,content,pos):
        self.type = type
        self.content = content
        self.pos = pos

def lex(expr):
    class State:
        NONE     = 0
        ID       = 1
        NUM      = 2
        NUM_AC   = 3
        NUM_EXP1 = 4
        NUM_EXP  = 5

    expr += " "

    state = State.NONE
    tok = ""

    i = 0
    while True:
        if state == State.NONE:
            if expr[i].isspace():
                i+=1
                if (i == len(expr)):
                    yield LexTok(LexTok.EOF,None,i)
                    return
            elif expr[i].isalpha() or expr[i] == "_":
                state = State.ID
            elif expr[i].isnumeric():
                state = State.NUM
            else:
                op = expr[i:][:2]
                if op in [">=","<=","==","!=","->"]:
                    yield LexTok(LexTok.SYM,op,i)
                    i+=2
                else:
                    op = expr[i:][:1]
                    if op in ["(",")","[","]","{","}","+","-",
                            "*","/","^","=","<",">",",","@","$"]:
                        yield LexTok(LexTok.SYM,op,i)
                        i+=1
                    else:
                        raise ParseError("Unknown character",i)
        if state == State.ID:
            if expr[i].isalnum() or expr[i] == "_":
                tok += expr[i]
                i += 1
            else:
                if tok in ["and","or","not","mod"]:
                    yield LexTok(LexTok.SYM,tok,i-len(tok))
                else:
                    yield LexTok(LexTok.ID,tok,i-len(tok))
                tok = ""
                state = State.NONE
        elif state >= State.NUM:
            if expr[i].isnumeric():
                tok += expr[i]
                i += 1
                if state == State.NUM_EXP1:
                    state = State.NUM_EXP
            elif state == State.NUM and expr[i] == ".":
                state = State.NUM_AC
                tok += expr[i]
                i += 1
            elif (state == State.NUM or state == State.NUM_AC) and (expr[i] == "e" or expr[i] == "E"):
                tok += expr[i]
                i += 1
                if expr[i] == "+" or expr[i] == "-":
                    tok += expr[i]
                    i += 1
                state = State.NUM_EXP1
            elif state == State.NUM_EXP1:
                raise ParseError("Expected digit in exponent",i)
            else:
                yield LexTok(LexTok.NUM,float(tok),i-len(tok))
                tok = ""
                state = State.NONE



from setuptools import setup

setup(name="gkraz",
        version="0.2",
        description="Calculator for the terminal",
        url="http://theovention.de",
        author="Theo Heimel",
        author_email="info@theovention.de",
        license="MIT",
        packages=["gkraz"],
        install_requires=[
            "prompt_toolkit",
            "pygments",
            "numpy",
            "matplotlib",
            "terminaltables",
            "pint"
        ],
        entry_points={
            "console_scripts": [
                "gkraz = gkraz.__main__:main"
            ]
        },
        zip_safe=False)



from collections import namedtuple, deque

from .lexer import *
from .bytecode import *
from .errors import ParseError

class BinaryOpAST:
    def __init__(self,op,left,right):
        self.op    = op
        self.left  = left
        self.right = right

    def gen_bc(self,bc):
        self.left.gen_bc(bc)
        self.right.gen_bc(bc)
        opcode = {
            "+":   OpCo.ADD,
            "-":   OpCo.SUB,
            "*":   OpCo.MUL,
            "/":   OpCo.DIV,
            "mod": OpCo.MOD,
            "and": OpCo.AND,
            "or":  OpCo.OR,
            "^":   OpCo.POW,
            "@":   OpCo.DOT,
            "->":  OpCo.TO
        }[self.op]
        bc.append(Bytecode(opcode))

class UnaryOpAST:
    def __init__(self,op,expr):
        self.op   = op
        self.expr = expr

    def gen_bc(self,bc):
        self.expr.gen_bc(bc)
        if self.op != "+":
            opcode = {
                "-":   OpCo.NEG,
                "not": OpCo.NOT,
            }[self.op]
            bc.append(Bytecode(opcode))

class CompOpAST:
    def __init__(self,ops,exprs):
        self.ops   = ops
        self.exprs = exprs

    def gen_bc(self,bc):
        for nr,expr in enumerate(self.exprs):
            expr.gen_bc(bc)
            if nr > 0:
                if nr < len(self.exprs)-1:
                    bc.append(Bytecode(OpCo.STO,0))
                opcode = {
                    ">":  OpCo.GT,
                    "<":  OpCo.LT,
                    ">=": OpCo.GTE,
                    "<=": OpCo.LTE,
                    "==": OpCo.EQ,
                    "!=": OpCo.NEQ
                }[self.ops[nr-1]]
                bc.append(Bytecode(opcode))
                if nr < len(self.exprs)-1:
                    bc.append(Bytecode(OpCo.LD,0))

        for i in range(1,len(self.ops)):
            bc.append(Bytecode(OpCo.AND))

class FuncCallAST:
    def __init__(self,name,params):
        self.name   = name
        self.params = params

    def gen_bc(self,bc):
        for param in self.params:
            param.gen_bc(bc)
        bc.append(Bytecode(OpCo.FUN,self.name,len(self.params)))

class VarExprAST:
    def __init__(self,name):
        self.name = name

    def gen_bc(self,bc):
        bc.append(Bytecode(OpCo.VAR,self.name))

class UnitAST:
    def __init__(self,unit):
        self.unit = unit

    def gen_bc(self,bc):
        bc.append(Bytecode(OpCo.UNI,self.unit))

class LiteralAST:
    def __init__(self,value):
        self.value = value

    def gen_bc(self,bc):
        bc.append(Bytecode(OpCo.LIT,self.value))

class VectorAST:
    def __init__(self,elements):
        self.elements = elements

    def gen_bc(self,bc):
        for element in self.elements:
            element.gen_bc(bc)
        bc.append(Bytecode(OpCo.VEC,len(self.elements)))

class FuncDefAST:
    def __init__(self,name,params,expr):
        self.name   = name
        self.params = params
        self.expr   = expr

    def gen_bc(self,bc):
        self.expr.gen_bc(bc)

class VarDefAST:
    def __init__(self,name,expr):
        self.name   = name
        self.expr   = expr

    def gen_bc(self,bc):
        self.expr.gen_bc(bc)

OpP = namedtuple("OpP",["ops","binary","comp"])

OP_PREC = [
    OpP( ["or"],                        True,  False ),
    OpP( ["and"],                       True,  False ),
    OpP( ["not"],                       False, False ),
    OpP( ["==","!=",">","<","<=",">="], True,  True  ),
    OpP( ["->"],                        True,  False ),
    OpP( ["+","-"],                     True,  False ),
    OpP( ["*","/","mod","@"],           True,  False ),
    OpP( ["^"],                         True,  False ),
    OpP( ["+","-"],                     False, False ),
]

class Parser:
    def __init__(self,tokens):
        self.tokens = tokens
        self.cur = next(self.tokens)
        self.la  = deque()

    def op_is(self,tok,op):
        return tok.type == LexTok.SYM and tok.content == op

    def accept(self,op):
        if self.op_is(self.cur,op):
            self.next_tok()
            return True
        else:
            return False

    def next_tok(self,n=1):
        for i in range(0,n):
            if len(self.la) == 0:
                self.cur = next(self.tokens)
            else:
                self.cur = self.la.popleft()

    def lookahead(self,n):
        if len(self.la) < n:
            for i in range(0,n-len(self.la)):
                self.la.append(next(self.tokens))
        return self.la[n-1]

    def parseStmt(self):
        stmt = self.parseVarDef() or self.parseFuncDef() or self.parseExpr()
        if self.cur.type != LexTok.EOF:
            raise ParseError("Expected end of statement",self.cur.pos)
        return stmt

    def parseVarDef(self):
        if self.cur.type != LexTok.ID:
            return None
        name = self.cur.content
        eq = self.lookahead(1)
        if not self.op_is(eq,"="):
            return None
        self.next_tok(2)

        expr = self.parseExpr()
        if expr is None:
            raise ParseError("Expected expression for variable value",self.cur.pos)
        return VarDefAST(name,expr)

    def parseFuncDef(self):
        if self.cur.type != LexTok.ID:
            return None
        name = self.cur.content
        bra = self.lookahead(1)
        if not self.op_is(bra,"("):
            return None
        params = []
        ntok = self.lookahead(2)
        lai = 2
        if ntok.type == LexTok.ID:
            params.append(ntok.content)
            lai += 1
            while self.op_is(self.lookahead(lai),","):
                lai += 1
                ntok = self.lookahead(lai)
                if ntok.type != LexTok.ID:
                    return None
                params.append(ntok.content)
                lai += 1
        if not self.op_is(self.lookahead(lai),")"):
            return None
        lai += 1
        if not self.op_is(self.lookahead(lai),"="):
            return None
        lai += 1

        self.next_tok(lai)
        expr = self.parseExpr()
        if expr is None:
            raise ParseError("Expected expression for function body",self.cur.pos)
        return FuncDefAST(name,params,expr)

    def parseExpr(self):
        return self.parseOpExpr(0)

    def parseFuncExpr(self):
        if self.cur.type != LexTok.ID:
            return None
        name = self.cur.content
        bra = self.lookahead(1)
        if not self.op_is(bra,"("):
            return None
        self.next_tok(2)

        params = []
        expr = self.parseExpr()
        if expr is not None:
            params.append(expr)
            while self.accept(","):
                expr = self.parseExpr()
                if expr is None:
                    raise ParseError("Expected parameter expression",self.cur.pos)
                params.append(expr)
        if not self.accept(")"):
            raise ParseError("Expected ')'",self.cur.pos)
        return FuncCallAST(name,params)

    def parseOpExpr(self,prec):
        if prec == len(OP_PREC):
            return self.parseMulExpr()
        opr = OP_PREC[prec]
        if opr.binary:
            expr1 = self.parseOpExpr(prec+1)
            if expr1 is None:
                return None
            if opr.comp:
                ops = []
                exprs = [expr1]

            while self.cur.type == LexTok.SYM and self.cur.content in opr.ops:
                op = self.cur.content
                self.next_tok()
                expr2 = self.parseOpExpr(prec+1)
                if expr2 is None:
                    raise ParseError("Expected expression after operator",self.cur.pos)
                if opr.comp:
                    ops.append(op)
                    exprs.append(expr2)
                else:
                    expr1 = BinaryOpAST(op,expr1,expr2)

            if opr.comp:
                if len(exprs) == 1:
                    return exprs[0]
                else:
                    return CompOpAST(ops,exprs)
            else:
                return expr1
        else:
            if self.cur.type == LexTok.SYM and self.cur.content in opr.ops:
                op = self.cur.content
                self.next_tok()
            else:
                op = None
            expr = self.parseOpExpr(prec+1)
            if op is None:
                return expr
            else:
                if expr is None:
                    raise ParseError("Expected expression after operator",self.cur.pos)
                else:
                    return UnaryOpAST(op,expr)

    def parseMulExpr(self):
        retexpr = None
        while True:
            expr = self.parsePrimaryExpr()
            if expr is None:
                break
            if retexpr is None:
                retexpr = expr
            else:
                retexpr = BinaryOpAST("*",retexpr,expr)

        return retexpr

    def parsePrimaryExpr(self):
        expr = self.parseFuncExpr()
        if expr is not None:
            return expr

        if self.cur.type == LexTok.ID:
            name = self.cur.content
            self.next_tok()
            return VarExprAST(name)

        if self.cur.type == LexTok.NUM:
            value = self.cur.content
            self.next_tok()
            return LiteralAST(value)

        if self.accept("("):
            expr = self.parseExpr()
            if expr is None:
                raise ParseError("Expected expression after opening bracket",self.cur.pos)
            if not self.accept(")"):
                raise ParseError("Expected ')'",self.cur.pos)
            return expr

        if self.accept("$"):
            if self.cur.type != LexTok.ID:
                raise ParseError("Expected unit name after $",self.cur.pos)
            unit = self.cur.content
            self.next_tok()
            return UnitAST(unit)

        if self.accept("["):
            elements = []
            expr = self.parseExpr()
            if expr is not None:
                elements.append(expr)
                while self.accept(","):
                    expr = self.parseExpr()
                    if expr is None:
                        raise ParseError("Expected expression after ','",self.cur.pos)
                    elements.append(expr)
            if not self.accept("]"):
                raise ParseError("Expected ']'",self.cur.pos)
            return VectorAST(elements)

        return None
